/*
 * serial_usb_servo - For managing servo motors on Arduino Mega 2560 using commands
 * from the USB serial.
 */
#include <Servo.h>

#define MAXNUMSERVOS 10
typedef struct {
  Servo servo;
  int id;   // unique identifier (0-99)
  int pin;  // Arduino pin
  int min;  // minimum angle
  int max;  // maximum angle
} servo_motor_t;
servo_motor_t sv[MAXNUMSERVOS];
int numservos = 0;

#define MAXCMDLEN 20
int maxcmdlen = MAXCMDLEN;
int lastgoodmsgno = -1;

typedef struct {
  char inbuf[MAXCMDLEN+1]; // room for a command and a null terminator
  int curchar = 0;
  int reading = false;
  char startchar = '<';
  char endchar = '>';
} serial_usb_t;

serial_usb_t su;

Servo servo_horiz; //@@tmp

void setup() {
  su.curchar = 0;
  su.reading = false;
  su.startchar = '<';
  su.endchar = '>';
  Serial.begin(115200);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB
  }
}

void loop() {
  int len, msgno;
  char msgnotext[3];
  
  if ((len = io_get_cmd()) > 0) {
    /* 
     * su.inbuf is now a '\0' terminated string containing:
     * 1. 2 digit message number represented as two ASCII numbers
     * 2. the command
     */
    // send back an ACK consisting of "S>" + the command + "<E"
    //Serial.write("S>", 2);
    //Serial.write(su.inbuf, len);
    //Serial.write("<E", 2);
    //Serial.flush(); // this doesn't seem to help, recipient is still somethings getting
                    // previous data followed by this one

    strncpy(msgnotext, su.inbuf, 2);
    msgnotext[2] = '\0';
    msgno = atoi(msgnotext);
    if (lastgoodmsgno != msgno) {
      lastgoodmsgno = msgno;
      // process the command
      process_cmd(&su.inbuf[2]);
    }
  }
}

/*
 * charston - Convert up to n ASCII numeric digits in chars to an integer and
 * return it. It stops when it encounters a non-numeric digit. This does at 
 * most 10 digits.
 */
int charston(char *chars, int n)
{
  int i;
  char buf[11];

  for (i = 0; i < 10 && i < n && isdigit(chars[i]); i++) {
    buf[i] = chars[i];
  }
  buf[i] = '\0';
  return atoi(buf);
}

int find_servo(int id)
{
  int i;

  for (i = 0; i < numservos; i++)
    if (sv[i].id == id)
      return i;
  return -1;
}
  
/*
 * process_cmd
 * 
 * General form of the commands:
 * <2_character_cmd>=<value>
 * 
 * <id> is 2 characters that are meant to uniquely identify a servo.
 * 
 * Commands:
 * cr=<id>,<pin> - Create an instance of a servo. Identify it using the <id>. 
 *                 Control is using the Arduino <pin>.
 * ma=<id>,<angle> - Move the servo to the given angle.
 * ac=<id>,<acceleration>
 * dc=<id>,<decelation>
 * sp=<id>,<speed>
 * 
 * Note that the first two characters of su.inbuf are the message number.
 */
int process_cmd(char *cmd)
{
  int id;

  id = charston(&cmd[3], 2);
  
  if (cmd[0] == 'c' && cmd[1] == 'r') { // create an instance of a servo
    if (strlen(cmd) >= 7 && cmd[2] == '=' && cmd[5] == ',') {
      int s;
      
      if ((s = find_servo(id)) == -1) {
        // we don't already have that one so add a new one to the end of the list
        s = numservos;
        numservos++;
      } // else reuse the existing one (should happen only during code development)
      sv[s].id = id;
      sv[s].pin = atoi(&cmd[6]);
      sv[s].min = 85;
      sv[s].max = 95;
      sv[s].servo.attach(sv[s].pin);
      Serial.print("cr"); Serial.print(id); Serial.print(sv[s].pin); Serial.flush();
    } // else too few characters in cmd
    
  } else if (cmd[0] == 'm' && cmd[1] == 'a') { // move the servo to the given angle
    int angle, s;

    if ((s = find_servo(id)) != -1) {
      angle = atoi(&cmd[6]);
      Serial.print("ma"); Serial.print(sv[s].pin); Serial.print(angle); Serial.flush();
      sv[s].servo.write(angle);
    } // else bad id
  }
  return 0;
}

/*
 * io_get_cmd() - Get a command from the serial port. Each command is expected
 * to be wrapped inside:
 * 1. single character su.startchar (e.g. '<')
 * 2. 2 digit message number represented as two ASCII numbers
 * 3. the command
 * 4. single character su.endchar (e.g. '<')
 * Note that this function doesn't care what's between the su.startchar and su.endchar.
 * 
 * This routine returns:
 * 0: read a character from the serial port, not finished getting the command
 * >0: a command has been fully read and is available as a '\0' terminated string
 *     starting at su.inbuf[2]. su.inbuf[0] and su.inbuf[1] are the messsage number.
 * -1: an error occurred, the next call to this function will start over looking for
 *     a command, any data gathered so far is lost
 */
int io_get_cmd() {
  char c;
  int len;
  
  if (Serial.available() > 0) {
    c = Serial.read();
    if (c == su.startchar) {
      if (su.curchar > 0) {
        // we shouldn't be getting this in the middle of a command
        su.curchar = 0;
        su.reading = false;
        return -1;
      }
      su.curchar = 0;
      su.reading = true;
      return 0; // starting a command
    }
    if (su.reading) {
      if (c == su.endchar) {
        // reached end of a command
        if (su.curchar > 0) {
          su.inbuf[su.curchar] = '\0';
          len = su.curchar;
          su.curchar = 0;
          su.reading = false;
          return len; // we got something that may be a complete command
        } else {
          su.curchar = 0;
          su.reading = false;
          return -1; // we got a zero-length command, the sender will have to retry
        }
      } else {
        // got a character for the command
        if (su.curchar < maxcmdlen-1) {
          su.inbuf[su.curchar] = c;
          su.curchar++;
        } else {
          // if we take this character then there won't be room for the null terminator,
          // reject the whole command, the sender will have to retry
          su.curchar = 0;
          su.reading = false;
          return -1;
        }
      }
    } else {
      return -1; // first character wasn't '>' but it should have been
    }
    return 0;
  }
  return 0;
}
